/*
 * Copyright 2013 University of Chicago and Argonne National Laboratory
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License
 */

package exm.stc.common.lang;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Set;

import exm.stc.common.lang.Types.Type;

/**
 * This class is used to contain the relevant information about
 * each variable in a Swift program.  All member variables
 * in the Variable are final, which means that each variable
 * instance can be treated as an immutable value and
 * safely shared between multiple data structures
 *
 */
public class Var {
  private final Type type;
  private final String name;
  private final VarStorage storage;
  private final DefType defType;
  private final Var mapping;

  public static final String TMP_VAR_PREFIX = "__t";
  public static final String ALIAS_VAR_PREFIX = "__alias";
  public static final String STRUCT_FIELD_VAR_PREFIX = "__sf_";
  public static final String LOCAL_VALUE_VAR_PREFIX = "__v_";
  public static final String FILENAME_OF_PREFIX = "__filename_";
  /* Separate value prefix to avoid name clashes for optimizer 
   *    inserted variables */
  public static final String OPT_VALUE_VAR_PREFIX = "__ov_";
  public static final String LOOP_INDEX_VAR_PREFIX = "__i";
  public static final String GLOBAL_CONST_VAR_PREFIX = "__c";
  public static final String DEREF_COMPILER_VAR_PREFIX = "__dr_";
  public static final String LOOP_COND_PREFIX = "__xcond";
  public static final String OUTER_VAR_PREFIX = "__outer";

  public enum VarStorage {
    /** Reference stored in global stack frame */
    STACK,
    /** Reference stored only temporarily in control flow */
    TEMP,
    /** Reference to variable allocated elsewhere */
    ALIAS,
    /** Value only stored in interpreter */
    LOCAL,
    /** Global constant */
    GLOBAL_CONST,
  }

  /**
   * How the variable was defined (e.g. local vs argument)
   */
  public enum DefType
  {
    /** Local variable */
    LOCAL_USER,
    /** Compiler-generated local */
    LOCAL_COMPILER,
    /** Input argument */
    INARG,
    /** Output argument */
    OUTARG,
    /** Global constant generated by compiler */
    GLOBAL_CONST,
  }

  public Var(Type type)
  {
    this.type = type;
    this.name = "...";
    this.storage = VarStorage.ALIAS;
    this.defType = DefType.INARG;
    this.mapping = null;
  }

  public Var(Type type, String name,
                  VarStorage storage, DefType defType)
  {
    this.type = type;
    this.name = name;
    this.storage = storage;
    this.defType = defType;
    this.mapping = null;
  }
  
  public Var(Type type, String name,
      VarStorage storage, DefType defType, Var mapping) {
    this.type = type;
    this.name = name;
    this.storage = storage;
    this.defType = defType;
    assert(mapping == null || Types.isString(mapping.type()));
    this.mapping = mapping;
  }

  public String name() {
    return name;
  }

  public Type type() {
    return type;
  }

  public VarStorage storage() {
    return storage;
  }

  public DefType defType() {
    return defType;
  }

  public Var mapping() {
    return mapping;
  }

  public boolean isMapped() {
    return mapping != null;
  }



  public static String names(List<Var> list)
  {
    StringBuilder sb = new StringBuilder(list.size()*16);
    Iterator<Var> it = list.iterator();
    while (it.hasNext())
    {
      Var v = it.next();
      sb.append(v.name());
      if (it.hasNext())
        sb.append(' ');
    }
    return sb.toString();
  }

  public static List<String> nameList(Collection<Var> variables)
  {
    List<String> result = new ArrayList<String>(variables.size());
    nameFill(result, variables);
    return result;
  }
  
  public static Set<String> nameSet(Collection<Var> variables)
  {
    Set<String> result = new HashSet<String>(variables.size());
    nameFill(result, variables);
    return result;
  }
  
  private static void nameFill(Collection<String> names,
                                  Collection<Var> variables)
  {
    for (Var v : variables)
      names.add(v.name());
  }

  public static List<Type> extractTypes(List<Var> variables)
  {
    ArrayList<Type> result = new ArrayList<Type>(variables.size());
    for (Var v: variables) {
      result.add(v.type());
    }

    return result;
  }
  
  /**
   * Different by variable name
   * @param list
   * @param subtract
   * @return
   */
  public static List<Var> varListDiff(List<Var> list, List<Var> subtract) {
    Set<String> subSet = nameSet(subtract);
    ArrayList<Var> diff = new ArrayList<Var>();
    for (Var v: list) {
      if (!subSet.contains(v.name())) {
        diff.add(v);
      }
    }
    return diff;
  }
  
  public static class VarCount {
    public Var var;
    public int count;
    public VarCount(Var var, int count) {
      super();
      this.var = var;
      this.count = count;
    }
  }
  public static List<VarCount> countVars(List<Var> list) {
    ArrayList<Var> sorted = new ArrayList<Var>(list);
    Collections.sort(sorted, new Comparator<Var>() {
      @Override
      public int compare(Var v1, Var v2) {
        return v1.name().compareTo(v2.name());
      }
    });
    ArrayList<VarCount> res = new ArrayList<VarCount>();
    VarCount curr = null;
    for (Var v: sorted) {
      if (curr == null || !v.name().equals(curr.var.name())) {
        curr = new VarCount(v, 1);
        res.add(curr);
      } else {
        curr.count++;
      }
    }
    return res;
  }

  /**
   * Union of lists with one instance of each variable by name
   * included in result
   * @param list1
   * @param list2
   * @return
   */
  public static List<Var> varListUnion(List<Var> list1, List<Var> list2) {
    ArrayList<Var> res = new ArrayList<Var>();
    Set<String> present = new HashSet<String>();
    for (Var v: list1) {
      if (!present.contains(v.name())) {
        res.add(v);
      }
    }
    for (Var v: list2) {
      if (!present.contains(v.name())) {
        res.add(v);
      }
    }
    return res;
  }
  
  /**
   * Create dereferenced variable given a reference
   */
  public static Var createDerefTmp(Var ref, VarStorage storage) {
    assert(Types.isRef(ref.type()));
    Var res = new Var(ref.type().memberType(),
        DEREF_COMPILER_VAR_PREFIX + ref.name(),
        storage, DefType.LOCAL_COMPILER, 
        null);
    assert(Types.isRefTo(ref.type(), res.type()));
    return res;
  }

  @Override
  public String toString()
  {
    return type.toString()+':'+name;
  }

  /**
   * Find the first variable with a matching name in the provided collection
   * @param variables
   * @param name
   * @return
   */
  public static Var findByName(Collection<Var> variables,
      String name) {
    for (Var v: variables) {
      if (v.name().equals(name)) {
        return v;
      }
    }
    return null;
  }
}
